/**
  This code isprovided AS-IS for educational purpose.
*/
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "php.h"
#include "php_ini.h"
#include "ext/standard/info.h"
#include "ext/standard/php_smart_str.h"
#include "ext/mysqlnd/mysqlnd.h"
#include "ext/mysqlnd/mysqlnd_alloc.h"
#include "ext/mysqlnd/mysqlnd_wireprotocol.h"
#include "php_mysqlnd_query_to_json.h"

#include "ext/mysqlnd/mysqlnd_priv.h"

static int plugin_id;

/* {{{ php_mysqlnd_qtj_net_field_length
   Get next field's length */
unsigned long
php_mysqlnd_qtj_net_field_length(zend_uchar **packet)
{
	register zend_uchar *p= (zend_uchar *)*packet;

	if (*p < 251) {
		(*packet)++;
		return (unsigned long) *p;
	}

	switch (*p) {
		case 251:
			(*packet)++;
			return MYSQLND_NULL_LENGTH;
		case 252:
			(*packet) += 3;
			return (unsigned long) uint2korr(p+1);
		case 253:
			(*packet) += 4;
			return (unsigned long) uint3korr(p+1);
		default:
			(*packet) += 9;
			return (unsigned long) uint4korr(p+1);
	}
}
/* }}} */

/* {{{ json_escape_string_qtj
   Escape a string for JSON usage */
static void json_escape_string_qtj(smart_str *buf, char *s, int len TSRMLS_DC)
{
	/* TODO: Support escaping of unicode chars */
	int pos = 0;
	unsigned short us;
	unsigned short *utf16;

	if (len == 0) {
		smart_str_appendl(buf, "\"\"", 2);
		return;
	}

	smart_str_appendc(buf, '"');

	while (pos < len)
	{
		us = s[pos++];

		switch (us)
		{
			case '"':
				smart_str_appendl(buf, "\\\"", 2);
				break;

			case '\\':
				smart_str_appendl(buf, "\\\\", 2);
				break;

			case '\b':
				smart_str_appendl(buf, "\\b", 2);
				break;

			case '\f':
				smart_str_appendl(buf, "\\f", 2);
				break;

			case '\n':
				smart_str_appendl(buf, "\\n", 2);
				break;

			case '\r':
				smart_str_appendl(buf, "\\r", 2);
				break;

			case '\t':
				smart_str_appendl(buf, "\\t", 2);
				break;

			default:
				smart_str_appendc(buf, (unsigned char) us);
				break;
		}
	}
	smart_str_appendc(buf, '"');
}
/* }}} */

/* {{{ process_row()
   Process a single COM_QUERY row and append the data as JSON to a smart_str */
static int process_row(MYSQLND_MEMORY_POOL_CHUNK * row_buffer, smart_str *json, unsigned int field_count, MYSQLND_FIELD *fields_metadata TSRMLS_DC)
{
	unsigned int i;
	zend_bool last_field_was_string = FALSE;
	zval **current_field, **end_field, **start_field;
	zend_uchar *p = row_buffer->ptr;
	size_t data_size = row_buffer->app;
	zend_uchar *bit_area = (zend_uchar*) row_buffer->ptr + data_size + 1; /* we allocate from here */

	if (!json) {
		return FAILURE;
	}

	for (i = 0; i < field_count; i++) {
		zend_uchar *this_field_len_pos = p;
		unsigned long len = php_mysqlnd_qtj_net_field_length(&p);

		if (i) {
			smart_str_appendc(json, ',');
		}
		if (len == MYSQLND_NULL_LENGTH) {
			smart_str_appendl(json, "null", sizeof("null")-1);
		} else {
			if (fields_metadata[i].type == MYSQL_TYPE_BIT) {
				/* TODO */
			} else {
				json_escape_string_qtj(json, (char *)p, len TSRMLS_CC);
				p += len;
			}
		}
	}
	return SUCCESS;
}
/* }}} */

/* {{{ proto string mysqlnd_query_to_json(mixed $conn, string $query)
       Execute a query and return the result as JSON data*/
PHP_FUNCTION(mysqlnd_query_to_json)
{
	zval *conn_zv;
	MYSQLND *conn;
	char *query;
	int query_len;
	smart_str retval = {0};

	MYSQLND_RES *result;
    MYSQLND_PACKET_ROW  *row_packet;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "zs", &conn_zv, &query, &query_len) == FAILURE) {
		return;
	}

	if (!(conn = zval_to_mysqlnd(conn_zv))) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Passed variableis no mysqlnd based connection");
		RETURN_FALSE;
	}

	if (FAIL == mysqlnd_query(conn, query, query_len TSRMLS_CC)) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Failed to execute query: %s (%u)", mysqlnd_get_error_str(conn TSRMLS_CC), mysqlnd_get_error_no(conn TSRMLS_CC));
		RETURN_FALSE;
	}

	/* based on mysqlnd_fetch_row_unbuffered_c */

	result = mysqlnd_use_result(conn);
	row_packet = result->row_packet;

	if (CONN_GET_STATE(result->conn) != CONN_FETCHING_DATA) {
		SET_CLIENT_ERROR(result->conn->error_info, CR_COMMANDS_OUT_OF_SYNC,
				UNKNOWN_SQLSTATE, mysqlnd_out_of_sync);
		RETURN_FALSE;
	}

	row_packet->skip_extraction = FALSE;

	smart_str_appendc(&retval, '[');
	while (!result->unbuf->eof_reached) {
		unsigned int i, field_count;
		MYSQLND_FIELD *field = result->meta->fields;
		struct mysqlnd_field_hash_key * hash_key = result->meta->zend_hash_keys;

		row_packet->skip_extraction = FALSE;

		if (PASS != PACKET_READ(row_packet, result->conn) || row_packet->eof) {
			result->unbuf->eof_reached = TRUE;
			break;
		}

		result->m.unbuffered_free_last_data(result TSRMLS_CC);

		result->unbuf->last_row_data = row_packet->fields;
		result->unbuf->last_row_buffer = row_packet->row_buffer;
		row_packet->fields = NULL;
		row_packet->row_buffer = NULL;

		if (!row_packet->skip_extraction) {
			if (retval.len > 1) {
				smart_str_appendc(&retval, ',');
			}
			smart_str_appendc(&retval, '[');

			process_row(result->unbuf->last_row_buffer, &retval, row_packet->field_count, row_packet->fields_metadata TSRMLS_CC);
			smart_str_appendc(&retval, ']');
			smart_str_appendc(&retval, '\n');
			result->m.unbuffered_free_last_data(result TSRMLS_CC);
		}
	}
	smart_str_appendc(&retval, ']');
	smart_str_appendc(&retval, 0);

	mysqlnd_free_result(result, TRUE);

	RETVAL_STRINGL(retval.c, retval.len-1, 1);
	smart_str_free(&retval);
}
/* }}} */

/* {{{ mysqlnd_query_to_json_functions[]
 */
const zend_function_entry mysqlnd_query_to_json_functions[] = {
	PHP_FE(mysqlnd_query_to_json,	NULL)
	PHP_FE_END
};
/* }}} */

/* {{{ PHP_MINIT_FUNCTION
 */
PHP_MINIT_FUNCTION(mysqlnd_query_to_json)
{
	plugin_id = mysqlnd_plugin_register();
	return SUCCESS;
}
/* }}} */

/* {{{ PHP_MINFO_FUNCTION
 */
PHP_MINFO_FUNCTION(mysqlnd_query_to_json)
{
	php_info_print_table_start();
	php_info_print_table_header(2, "mysqlndto query support", "enabled");
	php_info_print_table_end();
}
/* }}} */

/* {{{ mysqlnd_query_to_json_module_entry
 *  */
zend_module_entry mysqlnd_query_to_json_module_entry = {
	STANDARD_MODULE_HEADER,
	"mysqlnd_query_to_json",
	mysqlnd_query_to_json_functions,
	PHP_MINIT(mysqlnd_query_to_json),
	NULL,
	NULL,
	NULL,
	PHP_MINFO(mysqlnd_query_to_json),
	"0.1",
	STANDARD_MODULE_PROPERTIES
};
/* }}} */

#ifdef COMPILE_DL_MYSQLND_QUERY_TO_JSON
ZEND_GET_MODULE(mysqlnd_query_to_json)
#endif

/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 * vim600: noet sw=4 ts=4 fdm=marker
 * vim<600: noet sw=4 ts=4
 */
